// -*- coding: iso-8859-1 -*-
/**
* API - SDK Openbus C++
* \file openbus/ConnectionManager.h
*/

#ifndef _TECGRAF_MANAGER_H_
#define _TECGRAF_MANAGER_H_

#include <CORBA.h>
#include <string>
#include <vector>

namespace openbus {
  class ConnectionManager;
}

#include "openbus/interceptors/ORBInitializer_impl.h"
#include "openbus/Connection.h"

#define CONNECTION_MANAGER_ID "OpenbusConnectionManager"

/**
* \brief openbus
*/
namespace openbus {
  /**
   * \brief Gerencia conexes de acesso a barramentos OpenBus atravs de um ORB.
   *
   * Conexes representam formas diferentes de acesso ao barramento. O ConnectionManager permite
   * criar essas conexes e gerenci-las, indicando quais so utilizadas em cada chamada. As
   * conexes so usadas basicamente de duas formas no tratamento das chamadas: 
   * - para realizar uma chamada remota (cliente), neste caso a conexo  denominada "Requester".
   * - para validar uma chamada recebida (servidor), neste caso a conexo  denominada "Dispatcher".
   */
  class ConnectionManager : public CORBA::LocalObject {
  public:
    /**
     * \brief Cria uma conexo para um barramento.
     * 
     * Cria uma conexo para um barramento. O barramento  indicado por um nome ou endereo de rede
     * e um nmero de porta, onde os servios ncleo daquele barramento esto executando.
     * 
     * @param[in] host Endereo ou nome de rede onde os servios ncleo do barramento esto 
     *            executando.
     * @param[in] port Porta onde os servios ncleo do barramento esto executando.
     * @param[in] props Lista opcional de propriedades que definem algumas
     *        configuraes sobre a forma que as chamadas realizadas ou validadas
     *        com essa conexo so feitas. A seguir so listadas as propriedades
     *        vlidas:
     *        - access.key: chave de acesso a ser utiliza internamente para a
     *          gerao de credenciais que identificam as chamadas atravs do
     *          barramento. A chave deve ser uma chave privada RSA de 2048 bits
     *          (256 bytes). Quando essa propriedade no  fornecida, uma chave
     *          de acesso  gerada automaticamente.
     *        - legacy.disable: desabilita o suporte a chamadas usando protocolo
     *          OpenBus 1.5. Por padro o suporte est habilitado.
     *        - legacy.delegate: indica como  preenchido o campo 'delegate' das
     *          credenciais enviadas em chamadas usando protocolo OpenBus 1.5. H
     *          duas formas possveis (o padro  'caller'):
     *          - caller: o campo 'delegate'  preenchido sempre com a entidade
     *            do campo 'caller' da cadeia de chamadas.
     *          - originator: o campo 'delegate'  preenchido sempre com a
     *            entidade que originou a cadeia de chamadas, que  o primeiro
     *            login do campo 'originators' ou o campo 'caller' quando este
     *             vazio.
     *   
     * @throw InvalidBusAddress Os parmetros 'host' e 'port' no so vlidos.
     * @throw InvalidPropertyValue O valor de uma propriedade no  vlido.
     * @throw CORBA::Exception
     *
     * @return Conexo criada.
     */
    std::auto_ptr<Connection> createConnection(const char *host, short port,
                                               std::vector<std::string> props 
                                               = std::vector<std::string>());
    
    /**
     * \brief Define a conexo padro a ser usada nas chamadas.
     * 
     * Define uma conexo a ser utilizada como "Requester" e "Dispatcher" de chamadas sempre que no
     * houver uma conexo "Requester" e "Dispatcher" especfica definida para o caso especfico,
     * como  feito atravs das operaes 'setRequester' e 'setDispatcher'.
     * 
     * @param[in] conn Conexo a ser definida como conexo padro. O 'ownership' da conexo no 
     * transferida para o ConnectionManager, e a conexo deve ser removida do ConnectionManager
     * antes de destruida
     */
    void setDefaultConnection(Connection *conn) { _defaultConnection = conn; }
    
    /**
     * \brief Devolve a conexo padro.
     * 
     * Veja operao 'setDefaultConnection'.
     * 
     * \return Conexo definida como conexo padro. ConnectionManager no possui ownership dessa
     * conexo e o mesmo no  transferido para o cdigo de usurio na execuo desta funo
     */
    Connection * getDefaultConnection() const { return _defaultConnection; }
    
    /**
     * \brief Define a conexo "Requester" do contexto corrente.
     * 
     * Define a conexo "Requester" a ser utilizada em todas as chamadas feitas no contexto
     * atual. Quando 'conn'  'null' o contexto passa a ficar sem nenhuma conexo associada.
     * 
     * @param[in] conn Conexo a ser associada ao contexto corrente. O 'ownership' da conexo no 
     * transferida para o ConnectionManager, e a conexo deve ser removida do ConnectionManager
     * antes de destruida
     */
    void setRequester(Connection *conn);
    
    /**
     * \brief Devolve a conexo associada ao contexto corrente.
     * 
     * @throw CORBA::Exception
     *
     * @return Conexo a barramento associada a thread corrente. ConnectionManager no possui
     * ownership dessa conexo e o mesmo no  transferido para o cdigo de usurio na execuo
     * desta funo
     */
    Connection *getRequester() const;
    
    /**
     * \brief Define uma a conexo como "Dispatcher" de barramento.
     * 
     * Define a conexo como "Dispatcher" do barramento ao qual ela est conectada, de forma que
     * todas as chamadas originadas por entidades conectadas a este barramento sero validadas com
     * essa conexo. S pode haver uma conexo "Dispatcher" para cada barramento, portanto se j
     * houver outra conexo "Dispatcher" para o mesmo barramento essa ser substituda pela nova
     * conexo.
     * 
     * @param[in] conn Conexo a ser definida como "Dispatcher". O 'ownership' da conexo no 
     * transferida para o ConnectionManager, e a conexo deve ser removida do ConnectionManager
     * antes de destruida
     */
    void setDispatcher(Connection &conn);
    
    /**
     * \brief Devolve a conexo "Dispatcher" do barramento indicado.
     * 
     * @param[in] busid Identificador do barramento ao qual a conexo est
     * associada. ConnectionManager no possui ownership dessa conexo e o mesmo no  transferido
     * para o cdigo de usurio na execuo desta funo
     * 
     * @return Conexo "Dispatcher" do barramento indicado, ou 'null' caso no haja nenhuma conexo
     *         "Dispatcher" associada ao barramento indicado. ConnectionManager no possui ownership
     *         dessa conexo e o mesmo no  transferido para o cdigo de usurio na execuo desta
     *         funo
     */
    Connection * getDispatcher(const char *busid);
    
    /**
     * \brief Remove a conexo "Dispatcher" associada ao barramento indicado.
     * 
     * \param busid Identificador do barramento ao qual a conexo est associada.
     * 
     * \return Conexo "Dispatcher" associada ao barramento ou 'null' se no houver nenhuma conexo
     *         associada. ConnectionManager no possui ownership dessa conexo e o mesmo no 
     *         transferido para o cdigo de usurio na execuo desta funo
     */
    
    /**
     * \brief Devolve a conexo "Dispatcher" do barramento indicado.
     * 
     * \param busid Identificador do barramento ao qual a conexo est associada. ConnectionManager
     * no possui ownership dessa conexo e o mesmo no  transferido para o cdigo de usurio na
     * execuo desta funo
     * 
     * \return Conexo "Dispatcher" do barramento indicado, ou 'null' caso no haja nenhuma conexo
     *         "Dispatcher" associada ao barramento indicado.
     */
    Connection * clearDispatcher(const char *busid);
    
    /** 
     * ORB utilizado pela conexo. 
     */
    CORBA::ORB * orb() const { return _orb; }
  private:
    /**
    * ConnectionManager deve ser adquirido atravs de:
    *   orb->resolve_initial_references(CONNECTION_MANAGER_ID)
    */
    ConnectionManager(CORBA::ORB *, IOP::Codec *, 
      PortableInterceptor::SlotId slotId_joinedCallChain, 
      PortableInterceptor::SlotId slotId_signedCallChain, 
      PortableInterceptor::SlotId slotId_legacyCallChain,
      PortableInterceptor::SlotId slotId_requesterConnection,
      PortableInterceptor::SlotId slotId_receiveConnection);
    ~ConnectionManager();
    void orb(CORBA::ORB *o) { _orb = o; }
    typedef std::map<std::string, Connection*> BusidConnection;
    Mutex _mutex;
    CORBA::ORB *_orb;
    PortableInterceptor::Current_var _piCurrent;
    IOP::Codec *_codec;
    PortableInterceptor::SlotId _slotId_joinedCallChain; 
    PortableInterceptor::SlotId _slotId_signedCallChain;
    PortableInterceptor::SlotId _slotId_legacyCallChain;
    PortableInterceptor::SlotId _slotId_requesterConnection;
    PortableInterceptor::SlotId _slotId_receiveConnection;
    Connection *_defaultConnection;
    BusidConnection _busidConnection;
    friend CORBA::ORB *openbus::ORBInitializer(int argc, char **argv);
  };
}

#endif
